# -*- coding: utf-8 -*-
""" MDF v2 and v3 constants """


# byte order
BYTE_ORDER_INTEL = 0
BYTE_ORDER_MOTOROLA = 1

# data types
DATA_TYPE_UNSIGNED = 0
DATA_TYPE_SIGNED = 1
DATA_TYPE_FLOAT = 2
DATA_TYPE_DOUBLE = 3
DATA_TYPE_STRING = 7
DATA_TYPE_BYTEARRAY = 8
DATA_TYPE_UNSIGNED_INTEL = 13
DATA_TYPE_UNSIGNED_MOTOROLA = 9
DATA_TYPE_SIGNED_INTEL = 14
DATA_TYPE_SIGNED_MOTOROLA = 10
DATA_TYPE_FLOAT_INTEL = 15
DATA_TYPE_FLOAT_MOTOROLA = 11
DATA_TYPE_DOUBLE_INTEL = 16
DATA_TYPE_DOUBLE_MOTOROLA = 12

SIGNED_INT = {
    DATA_TYPE_SIGNED,
    DATA_TYPE_SIGNED_INTEL,
    DATA_TYPE_SIGNED_MOTOROLA,
}
STANDARD_INT_SIZES = {8, 16, 32, 64}

SIGNAL_TYPE_SCALAR = 0
SIGNAL_TYPE_STRING = 1
SIGNAL_TYPE_CANOPEN = 2
SIGNAL_TYPE_STRUCTURE_COMPOSITION = 3
SIGNAL_TYPE_ARRAY = 4

INT_TYPES = {
    DATA_TYPE_UNSIGNED,
    DATA_TYPE_SIGNED,
    DATA_TYPE_UNSIGNED_INTEL,
    DATA_TYPE_UNSIGNED_MOTOROLA,
    DATA_TYPE_SIGNED_INTEL,
    DATA_TYPE_SIGNED_MOTOROLA,
}

# channel types
CHANNEL_TYPE_VALUE = 0
CHANNEL_TYPE_MASTER = 1

# channel conversion types
CONVERSION_TYPE_NONE = 65535
CONVERSION_TYPE_LINEAR = 0
CONVERSION_TYPE_TABI = 1
CONVERSION_TYPE_TAB = 2
CONVERSION_TYPE_POLY = 6
CONVERSION_TYPE_EXPO = 7
CONVERSION_TYPE_LOGH = 8
CONVERSION_TYPE_RAT = 9
CONVERSION_TYPE_FORMULA = 10
CONVERSION_TYPE_TABX = 11
CONVERSION_TYPE_RTABX = 12

RAT_CONV_TEXT = '(P1 * X**2 + P2 * X + P3) / (P4 * X**2 + P5 * X + P6)'
POLY_CONV_SHORT_TEXT = 'P4 * X / P1'
POLY_CONV_LONG_TEXT = '(P2 - (P4 * (X - P5 -P6))) / (P3* (X - P5 - P6) - P1)'

DEPENDENCY_TYPE_NONE = 0
DEPENDENCY_TYPE_VECTOR = 1
DEPENDENCY_TYPE_NDIM = 256

# flags
FLAG_PRECISION = 1
FLAG_PHY_RANGE_OK = 2
FLAG_VAL_RANGE_OK = 8

# channel source types
SOURCE_ECU = 2
SOURCE_VECTOR = 19

# bus types
BUS_TYPE_NONE = 0
BUS_TYPE_CAN = 2
BUS_TYPE_FLEXRAY = 5

# file IO seek types
SEEK_START = 0
SEEK_REL = 1
SEEK_END = 2

# blocks size
ID_BLOCK_SIZE = 64
HEADER_COMMON_SIZE = 164
HEADER_POST_320_EXTRA_SIZE = 44
CE_BLOCK_SIZE = 128
FH_BLOCK_SIZE = 56
DG_PRE_320_BLOCK_SIZE = 24
DG_POST_320_BLOCK_SIZE = 28
HD_BLOCK_SIZE = 104
CN_DISPLAYNAME_BLOCK_SIZE = 228
CN_SHORT_BLOCK_SIZE = 218
CN_LONGNAME_BLOCK_SIZE = 222
CG_PRE_330_BLOCK_SIZE = 26
CG_POST_330_BLOCK_SIZE = 30
DT_BLOCK_SIZE = 24
CC_COMMON_BLOCK_SIZE = 46
CC_COMMON_SHORT_SIZE = 42
CC_ALG_BLOCK_SIZE = 88
CC_LIN_BLOCK_SIZE = 62
CC_POLY_BLOCK_SIZE = 94
CC_EXPO_BLOCK_SIZE = 102
CC_FORMULA_BLOCK_SIZE = 304
SR_BLOCK_SIZE = 156

# max int values
MAX_UINT8 = (1 << 8) - 1
MAX_UINT16 = (1 << 16) - 1
MAX_UINT32 = (1 << 32) - 1
MAX_UINT64 = (1 << 64) - 1

# data location
LOCATION_ORIGINAL_FILE = 0
LOCATION_TEMPORARY_FILE = 1
LOCATION_MEMORY = 2

# blocks struct fmts and keys
ID_FMT = '<8s8s8s4H2s26s2H'
ID_KEYS = (
    'file_identification',
    'version_str',
    'program_identification',
    'byte_order',
    'float_format',
    'mdf_version',
    'code_page',
    'reserved0',
    'reserved1',
    'unfinalized_standard_flags',
    'unfinalized_custom_flags',
)

HEADER_COMMON_FMT = '<2sH3IH10s8s32s32s32s32s'
HEADER_COMMON_KEYS = (
    'id',
    'block_len',
    'first_dg_addr',
    'comment_addr',
    'program_addr',
    'dg_nr',
    'date',
    'time',
    'author',
    'department',
    'project',
    'subject',
)

HEADER_POST_320_EXTRA_FMT = 'Q2H32s'
HEADER_POST_320_EXTRA_KEYS = (
    'abs_time',
    'tz_offset',
    'time_quality',
    'timer_identification',
)

FMT_CHANNEL_DISPLAYNAME = '<2sH5IH32s128s4H3d2IH'
KEYS_CHANNEL_DISPLAYNAME = (
    'id',
    'block_len',
    'next_ch_addr',
    'conversion_addr',
    'source_addr',
    'ch_depend_addr',
    'comment_addr',
    'channel_type',
    'short_name',
    'description',
    'start_offset',
    'bit_count',
    'data_type',
    'range_flag',
    'min_raw_value',
    'max_raw_value',
    'sampling_rate',
    'long_name_addr',
    'display_name_addr',
    'aditional_byte_offset',
)

FMT_CHANNEL_SHORT = '<2sH5IH32s128s4H3d'
FMT_CHANNEL_LONGNAME = FMT_CHANNEL_SHORT + 'I'
KEYS_CHANNEL_SHORT = (
    'id',
    'block_len',
    'next_ch_addr',
    'conversion_addr',
    'source_addr',
    'ch_depend_addr',
    'comment_addr',
    'channel_type',
    'short_name',
    'description',
    'start_offset',
    'bit_count',
    'data_type',
    'range_flag',
    'min_raw_value',
    'max_raw_value',
    'sampling_rate',
)
KEYS_CHANNEL_LONGNAME = KEYS_CHANNEL_SHORT + ('long_name_addr',)

FMT_CHANNEL_GROUP = '<2sH3I3HI'
KEYS_CHANNEL_GROUP = (
    'id',
    'block_len',
    'next_cg_addr',
    'first_ch_addr',
    'comment_addr',
    'record_id',
    'ch_nr',
    'samples_byte_nr',
    'cycles_nr',
)

FMT_DATA_GROUP_POST_320 = '<2sH4I2H4s'
KEYS_DATA_GROUP_POST_320 = (
    'id',
    'block_len',
    'next_dg_addr',
    'first_cg_addr',
    'trigger_addr',
    'data_block_addr',
    'cg_nr',
    'record_id_len',
    'reserved0',
)

FMT_DATA_GROUP_PRE_320 = '<2sH4I2H'
KEYS_DATA_GROUP_PRE_320 = (
    'id',
    'block_len',
    'next_dg_addr',
    'first_cg_addr',
    'trigger_addr',
    'data_block_addr',
    'cg_nr',
    'record_id_len',
)

FMT_SOURCE_COMMON = '<2s2H'
FMT_SOURCE_ECU = '<2s3HI80s32s4s'
FMT_SOURCE_EXTRA_ECU = '<HI80s32s4s'
KEYS_SOURCE_ECU = (
    'id',
    'block_len',
    'type',
    'module_nr',
    'module_address',
    'description',
    'ECU_identification',
    'reserved0',
)

FMT_SOURCE_VECTOR = '<2s2H2I36s36s42s'
FMT_SOURCE_EXTRA_VECTOR = '<2I36s36s42s'
KEYS_SOURCE_VECTOR = (
    'id',
    'block_len',
    'type',
    'CAN_id',
    'CAN_ch_index',
    'message_name',
    'sender_name',
    'reserved0',
)

KEYS_TEXT_BLOCK = (
    'id',
    'block_len',
    'text',
)

FMT_CONVERSION_COMMON = FMT_CONVERSION_NONE = '<2s2H2d20s2H'
FMT_CONVERSION_COMMON_SHORT = '<H2d20s2H'

KEYS_CONVESION_NONE = (
    'id',
    'block_len',
    'range_flag',
    'min_phy_value',
    'max_phy_value',
    'unit',
    'conversion_type',
    'ref_param_nr',
)

FMT_CONVERSION_FORMULA = '<2s2H2d20s2H{}s'
KEYS_CONVESION_FORMULA = (
    'id',
    'block_len',
    'range_flag',
    'min_phy_value',
    'max_phy_value',
    'unit',
    'conversion_type',
    'ref_param_nr',
    'formula',
)

FMT_CONVERSION_LINEAR = '<2s2H2d20s2H2d'
KEYS_CONVESION_LINEAR = (
    'id',
    'block_len',
    'range_flag',
    'min_phy_value',
    'max_phy_value',
    'unit',
    'conversion_type',
    'ref_param_nr',
    'b',
    'a',
)

FMT_CONVERSION_POLY_RAT = '<2s2H2d20s2H6d'
KEYS_CONVESION_POLY_RAT = (
    'id',
    'block_len',
    'range_flag',
    'min_phy_value',
    'max_phy_value',
    'unit',
    'conversion_type',
    'ref_param_nr',
    'P1',
    'P2',
    'P3',
    'P4',
    'P5',
    'P6',
)

FMT_CONVERSION_EXPO_LOGH = '<2s2H2d20s2H7d'
KEYS_CONVESION_EXPO_LOGH = (
    'id',
    'block_len',
    'range_flag',
    'min_phy_value',
    'max_phy_value',
    'unit',
    'conversion_type',
    'ref_param_nr',
    'P1',
    'P2',
    'P3',
    'P4',
    'P5',
    'P6',
    'P7',
)

FMT_PROGRAM_BLOCK = '<2sH{}s'
KEYS_PROGRAM_BLOCK = (
    'id',
    'block_len',
    'data',
)

FMT_SAMPLE_REDUCTION_BLOCK = '<2sH3Id'
KEYS_SAMPLE_REDUCTION_BLOCK = (
    'id',
    'block_len',
    'next_sr_addr',
    'data_block_addr',
    'cycles_nr',
    'time_interval',
)



# -*- coding: utf-8 -*-
""" asammdf *Signal* class module for time correct signal processing """

import logging
from textwrap import fill

import numpy as np

from .utils import MdfException, extract_cncomment_xml
from . import v2_v3_blocks as v3b
from . import v4_constants as v4c
from . import v4_blocks as v4b

from .version import __version__

logger = logging.getLogger('asammdf')


class Signal(object):
    """
    The *Signal* represents a channel described by it's samples and timestamps.
    It can perform arithmetic operations against other *Signal* or numeric types.
    The operations are computed in respect to the timestamps (time correct).
    The non-float signals are not interpolated, instead the last value relative
    to the current timestamp is used.
    *samples*, *timstamps* and *name* are mandatory arguments.

    Parameters
    ----------
    samples : numpy.array | list | tuple
        signal samples
    timestamps : numpy.array | list | tuple
        signal timestamps
    unit : str
        signal unit
    name : str
        signal name
    conversion : dict | channel conversion block
        dict that contains extra conversion information about the signal ,
        default *None*
    comment : str
        signal comment, default ''
    raw : bool
        signal samples are raw values, with no physical conversion applied
    master_metadata : list
        master name and sync type
    display_name : str
        display name used by mdf version 3
    attachment : bytes, name
        channel attachment and name from MDF version 4

    """

    def __init__(self,
                 samples=None,
                 timestamps=None,
                 unit='',
                 name='',
                 conversion=None,
                 comment='',
                 raw=True,
                 master_metadata=None,
                 display_name='',
                 attachment=(),
                 source=None,
                 bit_count=None):

        if samples is None or timestamps is None or name == '':
            message = ('"samples", "timestamps" and "name" are mandatory '
                       'for Signal class __init__: samples={}\n'
                       'timestamps={}\nname={}')
            raise MdfException(message.format(samples, timestamps, name))
        else:
            if isinstance(samples, (list, tuple)):
                samples = np.array(samples)
            if isinstance(timestamps, (list, tuple)):
                timestamps = np.array(timestamps, dtype=np.float64)
            if not samples.shape[0] == timestamps.shape[0]:
                message = '{} samples and timestamps length mismatch ({} vs {})'
                message = message.format(
                    name,
                    samples.shape[0],
                    timestamps.shape[0],
                )
                logger.exception(message)
                raise MdfException(message)
            self.samples = samples
            self.timestamps = timestamps
            self.unit = unit
            self.name = name
            self.comment = comment
            self._plot_axis = None
            self.raw = raw
            self.master_metadata = master_metadata
            self.display_name = display_name
            self.attachment = attachment
            self.source = source
            if bit_count is None:
                self.bit_count = samples.dtype.itemsize * 8
            else:
                self.bit_count = bit_count

            if not isinstance(conversion, (v4b.ChannelConversion, v3b.ChannelConversion)):
                if conversion is None:
                    pass

                elif 'a' in conversion:
                    conversion['conversion_type'] = v4c.CONVERSION_TYPE_LIN
                    conversion = v4b.ChannelConversion(
                        **conversion
                    )

                elif 'formula' in conversion:
                    conversion['conversion_type'] = v4c.CONVERSION_TYPE_ALG
                    conversion = v4b.ChannelConversion(
                        **conversion
                    )

                elif all(
                        key in conversion
                        for key in ['P{}'.format(i) for i in range(1, 7)]):
                    conversion['conversion_type'] = v4c.CONVERSION_TYPE_RAT
                    conversion = v4b.ChannelConversion(
                        **conversion
                    )

                elif 'raw_0' in conversion and 'phys_0' in conversion:
                    conversion['conversion_type'] = v4c.CONVERSION_TYPE_TAB
                    nr = 0
                    while 'phys_{}'.format(nr) in conversion:
                        nr += 1
                    conversion['val_param_nr'] = nr * 2
                    conversion = v4b.ChannelConversion(
                        **conversion
                    )

                elif 'upper_0' in conversion and 'phys_0' in conversion:
                    conversion['conversion_type'] = v4c.CONVERSION_TYPE_RTAB
                    nr = 0
                    while 'phys_{}'.format(nr) in conversion:
                        nr += 1
                    conversion['val_param_nr'] = nr * 3 + 1
                    conversion = v4b.ChannelConversion(
                        **conversion
                    )

                elif 'val_0' in conversion and 'text_0' in conversion:
                    conversion['conversion_type'] = v4c.CONVERSION_TYPE_TABX
                    nr = 0
                    while 'text_{}'.format(nr) in conversion:
                        nr += 1
                    conversion['ref_param_nr'] = nr + 1
                    conversion = v4b.ChannelConversion(
                        **conversion
                    )

                elif 'upper_0' in conversion and 'text_0' in conversion:
                    conversion['conversion_type'] = v4c.CONVERSION_TYPE_RTABX
                    nr = 0
                    while 'text_{}'.format(nr) in conversion:
                        nr += 1
                    conversion['ref_param_nr'] = nr + 1
                    conversion = v4b.ChannelConversion(
                        **conversion
                    )

                else:
                    conversion = v4b.ChannelConversion(
                        conversion_type=v4c.CONVERSION_TYPE_NON
                    )

            self.conversion = conversion

    def __repr__(self):
        string = """<Signal {}:
\tsamples={}
\ttimestamps={}
\tunit="{}"
\tconversion={}
\tsource={}
\tcomment="{}"
\tmastermeta="{}"
\traw={}
\tdisplay_name={}>
"""
        return string.format(
            self.name,
            self.samples,
            self.timestamps,
            self.unit,
            self.conversion,
            self.source,
            self.comment,
            self.master_metadata,
            self.raw,
            self.display_name,
        )

    def plot(self):
        """ plot Signal samples """
        try:
            import matplotlib.pyplot as plt
            from mpl_toolkits.mplot3d import axes3d
            from matplotlib.widgets import Slider
        except ImportError:
            logging.warning("Signal plotting requires matplotlib")
            return

        if len(self.samples.shape) <= 1 and self.samples.dtype.names is None:
            fig = plt.figure()
            fig.canvas.set_window_title(self.name)
            fig.text(0.95, 0.05, 'asammdf {}'.format(__version__),
                     fontsize=8, color='red',
                     ha='right', va='top', alpha=0.5)

            name = self.name

            if self.comment:
                comment = self.comment.replace('$', '')
                comment = extract_cncomment_xml(comment)
                comment = fill(comment, 120).replace('\\n', ' ')

                title = '{}\n({})'.format(name, comment)
                plt.title(title)
            else:
                plt.title(name)
            try:
                if not self.master_metadata:
                    plt.xlabel('Time [s]')
                    plt.ylabel('[{}]'.format(self.unit))
                    plt.plot(self.timestamps, self.samples, 'b')
                    plt.plot(self.timestamps, self.samples, 'b.')
                    plt.grid(True)
                    plt.show()
                else:
                    master_name, sync_type = self.master_metadata
                    if sync_type in (0, 1):
                        plt.xlabel('{} [s]'.format(master_name))
                    elif sync_type == 2:
                        plt.xlabel('{} [deg]'.format(master_name))
                    elif sync_type == 3:
                        plt.xlabel('{} [m]'.format(master_name))
                    elif sync_type == 4:
                        plt.xlabel('{} [index]'.format(master_name))
                    plt.ylabel('[{}]'.format(self.unit))
                    plt.plot(self.timestamps, self.samples, 'b')
                    plt.plot(self.timestamps, self.samples, 'b.')
                    plt.grid(True)
                    plt.show()
            except ValueError:
                plt.close(fig)
        else:
            try:
                names = self.samples.dtype.names
                if self.samples.dtype.names is None or len(names) == 1:

                    if names:
                        samples = self.samples[names[0]]
                    else:
                        samples = self.samples

                    shape = samples.shape[1:]

                    fig = plt.figure()
                    fig.canvas.set_window_title(self.name)
                    fig.text(
                        0.95,
                        0.05,
                        'asammdf {}'.format(__version__),
                        fontsize=8,
                        color='red',
                        ha='right',
                        va='top',
                        alpha=0.5,
                    )

                    if self.comment:
                        comment = self.comment.replace('$', '')
                        plt.title('{}\n({})'.format(self.name, comment))
                    else:
                        plt.title(self.name)

                    ax = fig.add_subplot(111, projection='3d')

                    # Grab some test data.
                    X = np.array(range(shape[1]))
                    Y = np.array(range(shape[0]))
                    X, Y = np.meshgrid(X, Y)

                    Z = samples[0]

                    # Plot a basic wireframe.
                    self._plot_axis = ax.plot_wireframe(X, Y, Z,
                                                        rstride=1, cstride=1)

                    # Place Sliders on Graph
                    ax_a = plt.axes([0.25, 0.1, 0.65, 0.03])

                    # Create Sliders & Determine Range
                    sa = Slider(ax_a,
                                'Time [s]',
                                self.timestamps[0],
                                self.timestamps[-1],
                                valinit=self.timestamps[0])

                    def update(val):
                        self._plot_axis.remove()
                        idx = np.searchsorted(self.timestamps,
                                              sa.val,
                                              side='right')
                        Z = samples[idx-1]
                        self._plot_axis = ax.plot_wireframe(
                            X,
                            Y,
                            Z,
                            rstride=1,
                            cstride=1,
                        )
                        fig.canvas.draw_idle()

                    sa.on_changed(update)

                    plt.show()

                else:
                    fig = plt.figure()
                    fig.canvas.set_window_title(self.name)
                    fig.text(
                        0.95,
                        0.05,
                        'asammdf {}'.format(__version__),
                        fontsize=8,
                        color='red',
                        ha='right',
                        va='top',
                        alpha=0.5,
                    )

                    if self.comment:
                        comment = self.comment.replace('$', '')
                        plt.title('{}\n({})'.format(self.name, comment))
                    else:
                        plt.title(self.name)

                    ax = fig.add_subplot(111, projection='3d')

                    samples = self.samples[names[0]]
                    axis1 = self.samples[names[1]]
                    axis2 = self.samples[names[2]]

                    # Grab some test data.
                    X, Y = np.meshgrid(axis2[0], axis1[0])

                    Z = samples[0]

                    # Plot a basic wireframe.
                    self._plot_axis = ax.plot_wireframe(
                        X,
                        Y,
                        Z,
                        rstride=1,
                        cstride=1,
                    )

                    # Place Sliders on Graph
                    ax_a = plt.axes([0.25, 0.1, 0.65, 0.03])

                    # Create Sliders & Determine Range
                    sa = Slider(ax_a,
                                'Time [s]',
                                self.timestamps[0],
                                self.timestamps[-1],
                                valinit=self.timestamps[0])

                    def update(val):
                        self._plot_axis.remove()
                        idx = np.searchsorted(self.timestamps,
                                              sa.val,
                                              side='right')
                        Z = samples[idx-1]
                        X, Y = np.meshgrid(axis2[idx-1], axis1[idx-1])
                        self._plot_axis = ax.plot_wireframe(
                            X,
                            Y,
                            Z,
                            rstride=1,
                            cstride=1,
                        )
                        fig.canvas.draw_idle()

                    sa.on_changed(update)

                    plt.show()

            except Exception as err:
                print(err)

    def cut(self, start=None, stop=None):
        """
        Cuts the signal according to the *start* and *stop* values, by using
        the insertion indexes in the signal's *time* axis.

        Parameters
        ----------
        start : float
            start timestamp for cutting
        stop : float
            stop timestamp for cutting

        Returns
        -------
        result : Signal
            new *Signal* cut from the original

        Examples
        --------
        >>> new_sig = old_sig.cut(1.0, 10.5)
        >>> new_sig.timestamps[0], new_sig.timestamps[-1]
        0.98, 10.48

        """
        if len(self) == 0:
            result = Signal(
                np.array([]),
                np.array([]),
                self.unit,
                self.name,
                self.conversion,
                self.comment,
                self.raw,
                self.master_metadata,
                self.display_name,
                self.attachment,
            )

        elif start is None and stop is None:
            # return the channel uncut
            result = Signal(
                self.samples.copy(),
                self.timestamps.copy(),
                self.unit,
                self.name,
                self.conversion,
                self.comment,
                self.raw,
                self.master_metadata,
                self.display_name,
                self.attachment,
            )

        else:
            if start is None:
                # cut from begining to stop
                stop = np.searchsorted(self.timestamps, stop, side='right')
                if stop:
                    result = Signal(
                        self.samples[:stop],
                        self.timestamps[:stop],
                        self.unit,
                        self.name,
                        self.conversion,
                        self.comment,
                        self.raw,
                        self.master_metadata,
                        self.display_name,
                        self.attachment,
                    )
                else:
                    result = Signal(
                        np.array([]),
                        np.array([]),
                        self.unit,
                        self.name,
                        self.conversion,
                        self.comment,
                        self.raw,
                        self.master_metadata,
                        self.display_name,
                        self.attachment,
                    )

            elif stop is None:
                # cut from start to end
                start = np.searchsorted(self.timestamps, start, side='left')
                result = Signal(
                    self.samples[start:],
                    self.timestamps[start:],
                    self.unit,
                    self.name,
                    self.conversion,
                    self.comment,
                    self.raw,
                    self.master_metadata,
                    self.display_name,
                    self.attachment,
                )

            else:
                # cut between start and stop
                if start > self.timestamps[-1]:
                    result = Signal(
                        np.array([]),
                        np.array([]),
                        self.unit,
                        self.name,
                        self.conversion,
                        self.comment,
                        self.raw,
                        self.master_metadata,
                        self.display_name,
                        self.attachment,
                    )
                else:
                    start_ = np.searchsorted(self.timestamps, start, side='left')
                    start_ = max(0, start_)
                    stop_ = np.searchsorted(self.timestamps, stop, side='right')

                    if start not in self.timestamps and start_ == stop_:
                        start_ -= 1
                    if stop_ == start_:
                        if (len(self.timestamps)
                                and stop >= self.timestamps[0]
                                and start <= self.timestamps[-1]):
                            # start and stop are found between 2 signal samples
                            # so return the previous sample
                            result = Signal(
                                self.samples[start_ - 1: start_],
                                self.timestamps[start_ - 1: start_],
                                self.unit,
                                self.name,
                                self.conversion,
                                self.comment,
                                self.raw,
                                self.master_metadata,
                                self.display_name,
                                self.attachment,
                            )
                        else:
                            # signal is empty or start and stop are outside the
                            # signal time base
                            result = Signal(
                                np.array([]),
                                np.array([]),
                                self.unit,
                                self.name,
                                self.conversion,
                                self.comment,
                                self.raw,
                                self.master_metadata,
                                self.display_name,
                                self.attachment,
                            )
                    else:
                        result = Signal(
                            self.samples[start_: stop_],
                            self.timestamps[start_: stop_],
                            self.unit,
                            self.name,
                            self.conversion,
                            self.comment,
                            self.raw,
                            self.master_metadata,
                            self.display_name,
                            self.attachment,
                        )
        return result

    def extend(self, other):
        """ extend signal with samples from another signal

        Parameters
        ----------
        other : Signal

        Returns
        -------
        signal : Signal
            new extended *Signal*

        """
        if len(self.timestamps):
            last_stamp = self.timestamps[-1]
            delta = last_stamp / len(self) + last_stamp
        else:
            last_stamp = 0
            delta = 0
        if len(other):
            other_first_sample = other.timestamps[0]
            if last_stamp >= other_first_sample:
                timestamps = other.timestamps + delta - other_first_sample
            else:
                timestamps = other.timestamps

            result = Signal(
                np.append(self.samples, other.samples),
                np.append(self.timestamps, timestamps),
                self.unit,
                self.name,
                self.conversion,
                self.comment,
                self.raw,
                self.master_metadata,
                self.display_name,
                self.attachment,
            )
        else:
            result = self

        return result

    def interp(self, new_timestamps):
        """ returns a new *Signal* interpolated using the *new_timestamps*

        Parameters
        ----------
        new_timestamps : np.array
            timestamps used for interpolation

        Returns
        -------
        signal : Signal
            new interpolated *Signal*

        """
        if not len(self.samples) or not len(new_timestamps):
            return Signal(
                self.samples.copy(),
                self.timestamps.copy(),
                self.unit,
                self.name,
                comment=self.comment,
                conversion=self.conversion,
                raw=self.raw,
                master_metadata=self.master_metadata,
                display_name=self.display_name,
                attachment=self.attachment,
            )
        else:
            if self.samples.dtype.kind == 'f':
                s = np.interp(new_timestamps, self.timestamps, self.samples)
            else:
                idx = np.searchsorted(
                    self.timestamps,
                    new_timestamps,
                    side='right',
                )
                idx -= 1
                idx = np.clip(idx, 0, idx[-1])
                s = self.samples[idx]

            return Signal(
                s,
                new_timestamps,
                self.unit,
                self.name,
                comment=self.comment,
                conversion=self.conversion,
                raw=self.raw,
                master_metadata=self.master_metadata,
                display_name=self.display_name,
                attachment=self.attachment,
            )

    def __apply_func(self, other, func_name):
        """ delegate operations to the *samples* attribute, but in a time
        correct manner by considering the *timestamps*

        """

        if isinstance(other, Signal):
            time = np.union1d(self.timestamps, other.timestamps)
            s = self.interp(time).samples
            o = other.interp(time).samples
            func = getattr(s, func_name)
            s = func(o)
        elif other is None:
            s = self.samples
            time = self.timestamps
        else:
            func = getattr(self.samples, func_name)
            s = func(other)
            time = self.timestamps
        return Signal(
            s,
            time,
            self.unit,
            self.name,
            self.conversion,
            self.raw,
            self.master_metadata,
            self.display_name,
            attachment=self.attachment,
        )

    def __pos__(self):
        return self

    def __neg__(self):
        return Signal(
            np.negative(self.samples),
            self.timestamps,
            self.unit,
            self.name,
            self.conversion,
            self.raw,
            self.master_metadata,
            self.display_name,
            self.attachment,
        )

    def __round__(self, n):
        return Signal(
            np.around(self.samples, n),
            self.timestamps,
            self.unit,
            self.name,
            self.conversion,
            self.raw,
            self.master_metadata,
            self.display_name,
            self.attachment,
        )

    def __sub__(self, other):
        return self.__apply_func(other, '__sub__')

    def __isub__(self, other):
        return self.__sub__(other)

    def __rsub__(self, other):
        return -self.__sub__(other)

    def __add__(self, other):
        return self.__apply_func(other, '__add__')

    def __iadd__(self, other):
        return self.__add__(other)

    def __radd__(self, other):
        return self.__add__(other)

    def __mul__(self, other):
        return self.__apply_func(other, '__mul__')

    def __imul__(self, other):
        return self.__mul__(other)

    def __rmul__(self, other):
        return self.__mul__(other)

    def __truediv__(self, other):
        return self.__apply_func(other, '__truediv__')

    def __itruediv__(self, other):
        return self.__truediv__(other)

    def __rtruediv__(self, other):
        return self.__apply_func(other, '__rtruediv__')

    def __mod__(self, other):
        return self.__apply_func(other, '__mod__')

    def __pow__(self, other):
        return self.__apply_func(other, '__pow__')

    def __and__(self, other):
        return self.__apply_func(other, '__and__')

    def __or__(self, other):
        return self.__apply_func(other, '__or__')

    def __xor__(self, other):
        return self.__apply_func(other, '__xor__')

    def __invert__(self):
        s = ~self.samples
        time = self.timestamps
        return Signal(
            s,
            time,
            self.unit,
            self.name,
            self.conversion,
            self.raw,
            self.master_metadata,
            self.display_name,
            self.attachment,
        )

    def __lshift__(self, other):
        return self.__apply_func(other, '__lshift__')

    def __rshift__(self, other):
        return self.__apply_func(other, '__rshift__')

    def __lt__(self, other):
        return self.__apply_func(other, '__lt__')

    def __le__(self, other):
        return self.__apply_func(other, '__le__')

    def __gt__(self, other):
        return self.__apply_func(other, '__gt__')

    def __ge__(self, other):
        return self.__apply_func(other, '__ge__')

    def __eq__(self, other):
        return self.__apply_func(other, '__eq__')

    def __ne__(self, other):
        return self.__apply_func(other, '__ne__')

    def __iter__(self):
        for item in (
                self.samples,
                self.timestamps,
                self.unit,
                self.name):
            yield item

    def __reversed__(self):
        return enumerate(zip(reversed(self.samples),
                             reversed(self.timestamps)))

    def __len__(self):
        return len(self.samples)

    def __abs__(self):
        return Signal(
            np.fabs(self.samples),
            self.timestamps,
            self.unit,
            self.name,
            self.conversion,
            self.raw,
            self.master_metadata,
            self.display_name,
            self.attachment,
        )

    def __getitem__(self, val):
        return self.samples[val]

    def __setitem__(self, idx, val):
        self.samples[idx] = val

    def astype(self, np_type):
        """ returns new *Signal* with samples of dtype *np_type*

        Parameters
        ----------
        np_type : np.dtype
            new numpy dtye

        Returns
        -------
        signal : Signal
            new *Signal* with the samples of *np_type* dtype

        """
        return Signal(
            self.samples.astype(np_type),
            self.timestamps,
            self.unit,
            self.name,
            self.conversion,
            self.raw,
            self.master_metadata,
            self.display_name,
            self.attachment,
        )

    def physical(self):
        """
        get the physical samples values

        Returns
        -------
        phys : Signal
            new *Signal* with physical values

        """

        if not self.raw or self.conversion is None:
            samples = self.samples.copy()
        else:
            samples = self.conversion.convert(self.samples)

        return Signal(
            samples,
            self.timestamps.copy(),
            unit=self.unit,
            name=self.name,
            raw=False,
            comment=self.comment,
            master_metadata=self.master_metadata,
            display_name=self.display_name,
            attachment=self.attachment,
        )


if __name__ == '__main__':
    pass
